/*
 *  The Regina Rexx Interpreter
 *  Copyright (C) 1992-1994  Anders Christensen <anders@pvv.unit.no>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: configur.h,v 1.11 1993/05/10 06:08:55 anders Exp anders $
 */

/*
 * For platforms that don't use autoconf, put these here...
 */
#if defined(DOS) && defined(EMX)
# define HAVE_ASSERT_H
# define HAVE_SETJMP_H
# define HAVE_UNISTD_H
# define HAVE_FULLPATH
# define HAVE_PUTENV
# define HAVE_GETTIMEOFDAY
# define HAVE_FTRUNCATE
# define HAVE_MEMCPY
# define HAVE_MEMMOVE
# define HAVE_STRERROR
# define HAVE_VFPRINTF
# define TIME_WITH_SYS_TIME
#endif

#if defined(OS2) && defined(EMX)
# define HAVE_ASSERT_H
# define HAVE_SETJMP_H
# define HAVE_UNISTD_H
# define HAVE_FULLPATH
# define HAVE_PUTENV
# define HAVE_GETTIMEOFDAY
# define HAVE_FTRUNCATE
# define HAVE_MEMCPY
# define HAVE_MEMMOVE
# define HAVE_STRERROR
# define HAVE_VFPRINTF
# define TIME_WITH_SYS_TIME
#endif

#if defined(OS2) && defined(__WATCOMC__)
# define HAVE_ASSERT_H
# define HAVE_MALLOC_H
# define HAVE_SETJMP_H
# define HAVE_UNISTD_H
# define HAVE__FULLPATH
# define HAVE_PUTENV
# define HAVE_MEMCPY
# define HAVE_MEMMOVE
# define HAVE_STRERROR
# define HAVE_VFPRINTF
# if defined(DYNAMIC)
#  define DYNAMIC_OS2
# endif
#endif

#if defined(WIN32) && defined(__WATCOMC__)
# define HAVE_ASSERT_H
# define HAVE_MALLOC_H
# define HAVE_SETJMP_H
# define HAVE_UNISTD_H
# define HAVE_FTIME
# define HAVE__FULLPATH
# define HAVE_MEMCPY
# define HAVE_MEMMOVE
# define HAVE_PUTENV
# define HAVE_STRERROR
# define HAVE_VFPRINTF
# if defined(DYNAMIC)
#  define DYNAMIC_WIN32
# endif
#endif

#if defined(WIN32) && defined(_MSC_VER)
# define HAVE_ASSERT_H
# define HAVE_MALLOC_H
# define HAVE_SETJMP_H
# define HAVE__FULLPATH
# define HAVE_PUTENV
# define HAVE_FTIME
# define HAVE_MEMCPY
# define HAVE_MEMMOVE
# define HAVE_STRERROR
# define HAVE_VFPRINTF
# if defined(DYNAMIC)
#  define DYNAMIC_WIN32
# endif
#endif

#if defined(DOS) && defined(DJGPP)
# define HAVE_ASSERT_H
# define HAVE_LIMITS_H
# define TIME_WITH_SYS_TIME
# define HAVE_SETJMP_H
# define HAVE_UNISTD_H
# define HAVE_PUTENV
# define HAVE_GETTIMEOFDAY
# define HAVE_FTRUNCATE
# define HAVE_RANDOM
# define HAVE_USLEEP
# define HAVE_MEMCPY
# define HAVE_MEMMOVE
# define HAVE_STRERROR
# define HAVE_TRUENAME
# define HAVE_VFPRINTF
# ifndef _MAX_PATH
#  define _MAX_PATH  PATH_MAX
# endif
#endif


/* Configurable values */


/* The macro MAXLEVELS is the maximum number of levels that the 
 *   interpreter shall handle. One level is used for functioncalls and
 *   the commands CALL, IF, DO and SELECT
 */
#define MAXLEVELS        1024

/* The macro HASHTABLENGTH is the number of sloths in the hash table
 *   containing the values of the rexx variables
 */
#define HASHTABLENGTH     256

/* The macro MAX_INDEX_LENGTH defines the maximum length that the name
 *   of an compund variable might expand to.
 */
#define MAX_INDEX_LENGTH  256

/*
 * The macro DEFAULT_FLUSH_STACK can be defined to either 1 or 0,
 * depending on whether you want the stack to be flushed during a
 * command reading input from or writing output to the stack. If
 * defined, a line written will be physically in the stack at once,
 * while if not defined, all output lines will be temporarily stacked
 * until the command has terminated, then they are all simultaneously
 * flushed to the stack. The default default value is 0, it can be 
 * changed through OPTIONS.
 */
#ifndef DEFAULT_FLUSHSTACK
# define DEFAULT_FLUSHSTACK 0
#endif

/*
 * The macro DEFAULT_LINEOUTTRUNC can be set to 1 or 0, whether you want
 * the default behavior of lineout() to truncate the file immediately 
 * after the end of each new file written out. When set to 1, the file
 * will be truncated. Also note, that if HAVE_FTRUNCATE isn't set later
 * in this script, the DEFAULT_LINEOUTTRUNC will be forced to 0, just 
 * for consistency (then, setting it to 1 will have no practical effect.
 */
#ifndef DEFAULT_LINEOUTTRUNC
# define DEFAULT_LINEOUTTRUNC 1
#endif

/*
 * Whether the OPEN() and CLOSE() bifs are available by default
 */
#define DEFAULT_OPEN_BIF 1
#define DEFAULT_CLOSE_BIF 1

#define DEFAULT_MAKEBUF_BIF 1
#define DEFAULT_DROPBUF_BIF 1
#define DEFAULT_BUFTYPE_BIF 1
#define DEFAULT_DESBUF_BIF 1

#define DEFAULT_CACHEEXT 0
#define DEFAULT_PRUNE_TRACE 1

#define DEFAULT_FIND_BIF 1 

/* 
 * Define the following if your machine has putenv(), unfortunately, 
 *   Standard C does not include a way to set environmentvariables, just 
 *   to read them. But most machines have the putenv() function.
 * If it is not defined, you will not be able to set the environment 
 *   variables throught the VALUE() builtin function. 
 */
/* #define HAVE_PUTENV */ /* handled in makefile - TBD in autoconf */

/* 
 * The FILE_SEPARATOR is one character, and defines the char that separates
 * directories, and filename from directories in a file's path.
 * The PATH_SEPARATOR is one character, and defines the char that separates
 * directories, from one another in a PATH environment variable.
 */

#if defined(MSDOS) || defined(__WATCOMC__) || defined(_MSC_VER) || defined(DOS) || defined(OS2)
# define FILE_SEPARATOR      '\\'
# define FILE_SEPARATOR_STR  "\\"
# define PATH_SEPARATOR      ';'
# define PATH_SEPARATOR_STR  ";"
#else
# if defined(VMS)
#  define FILE_SEPARATOR     ']'
#  define FILE_SEPARATOR_STR "]"
#  define PATH_SEPARATOR     '?'
#  define PATH_SEPARATOR_STR "?"
# else
#  define FILE_SEPARATOR     '/'
#  define FILE_SEPARATOR_STR "/"
#  define PATH_SEPARATOR     ':'
#  define PATH_SEPARATOR_STR ":"
# endif
#endif


#define FIX_PROTOS
#define FIX_ALL_PROTOS


extern unsigned char u_to_l[] ;
#define RXTOLOW(a) (u_to_l[(unsigned char)(a)])


#if defined(HPUX) || defined(__hpux)
# define YY_CHAR_TYPE unsigned char
#endif

#if !defined(HAVE_MEMCPY)
# define memcpy(a,b,c)  bcopy((b),(a),(c))
#endif

#if !defined(HAVE_MEMMOVE)
# define memmove(a,b,c) bcopy((b),(a),(c))
#endif

#if !defined(HAVE_STRERROR)
  char *get_sys_errlist( int num );
# define strerror(a)    get_sys_errlist(a)
#endif


#if defined(sgi) || defined(__sgi)
# define SGI 1
#endif

#if defined(__osf__)
# define OSF 1
# define YY_CHAR_TYPE unsigned char
#endif

/* 
 * Define HAS_USLEEP if you have the usleep() function available. It will
 *    take the number of micro-seconds to sleep. If this is not defined, 
 *    the 'normal' sleep() will be used.
 * Handled by autoconf
 */

/*
 * Define HAS_RANDOM if your machine has the routines random() and 
 *    srandom(). If it is not defined, rand() and srand() are used
 *    instead. There are several reason why you might want to use 
 *    random() instead of rand()
 * Handled by autoconf
 */


#define DEFAULT_NUMERIC_FUZZ 0
#define DEFAULT_NUMERIC_SIZE 9
#define DEFAULT_MATH_TYPE 0


/* 
 * If you use gcc on a machine that does not have proper ANSI C 
 * libraries and the function div() is not defined, then you can 
 * define the symbol HAS_NO_DIV, which will try to kludge it. 
 *
 * This is necessary on Suns, and some other machines.
 *
 * Not used anymore MH 10-06-97
 */


#ifndef YY_CHAR_TYPE
# ifdef _AIX
#  define YY_CHAR_TYPE unsigned char
# else
#  define YY_CHAR_TYPE char
# endif
#endif


/* Unfortunately, ftruncate() is not POSIX, so if you don't have it,
 * just ignore it, the effect is that lines written to the middle of a
 * file will not cause the rest of the file (after the linefeed
 * terminating the line written), also to truncate the file at that
 * point. If you wonder whether you have ftruncate(), the bsd portions
 * of you system (if it exist) is a good place to start looking. You
 * may have to link in a specific library too, maybe libbsd.a.
 */
/*#define HAVE_FTRUNCATE*/    /* handled in makefile - TBD in autoconf */



#ifdef VAXC
# ifdef HAVE_GETTIMEOFDAY
#  undef HAVE_GETTIMEOFDAY
# endif
# ifdef HAVE_PUTENV
#  undef HAVE_PUTENV
# endif
# ifdef HAS_FTRUNCATE
#  undef HAS_FTRUNCATE
# endif
#endif


#if defined(__GNUC__) && defined(VMS)
# ifdef HAVE_GETTIMEOFODAY
#  undef HAVE_GETTIMEOFDAY
# endif
#endif

#define OLD_REGINA_FEATURES

/*
 * DLOPEN   - dlopen()/dlsym() under SunOS and SysV4
 * HPSHLOAD - shl_load()/shl_findsym() under HP-UX
 * AIXLOAD  - load()/loadbind() under AIX - not supported
 * OS2      - DosLoadModule()/DosQueryProcAddr() under OS/2
 * WIN32    - LoadLibrary()/GetProcAddress() under Win32
 */

/*
 * If we don't have ftruncate() then we don't want the default setting
 * of LINEOUT() to be to truncate after each new line written out.
 */
#ifndef HAS_FTRUNCATE
# undef DEFAULT_LINEOUTTRUNC                            /* MH 10-06-96 */
# define DEFAULT_LINEOUTTRUNC 0
#endif



